/*
 * auth_clnt.c	This module takes care of request authorization.
 *
 * three top-level functions are provided:
 * - nfs_client *auth_clnt(struct svc_req *rqstp)
 *    Look up and return a client descriptor based on received request;
 *    return descriptor (or NULL if not authorized)
 * - nfs_mount *auth_path(nfs_client *cp, struct svc_req *rqstp, char *path)
 *    Look up mount descriptor; check if client "cp" is authorized to mount
 *    "path"; return nfs_mount (NULL if unauthorized). Used by mountd.
 * - int auth_user(nfs_mount *mp, struct svc_req *rqstp)
 *    Make sure uid/gid credentials that come with reequest *rqstp can
 *    access mount point *mp; if authorized, return 1, else 0
 *
 * Renato J. Figueiredo, based on original UNFSD code by:
 *
 *
 * Authors:	Don Becker, <becker@super.org>
 *		Rick Sladkey, <jrs@world.std.com>
 *		Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *		Olaf Kirch, <okir@monad.swb.de>
 *
 *		This software maybe be used for any purpose provided
 *		the above copyright notice is retained.  It is supplied
 *		as is, with no warranty expressed or implied.
 */


#include "system.h"
#include "nfsd.h"

#ifdef SOLARIS
#define PORTMAP
#include <rpc/rpc.h>
#include <rpc/svc_soc.h>
#endif /* SOLARIS */

#ifndef svc_getcaller
#define svc_getcaller(x) ((struct sockaddr_in *) (x)->xp_rtaddr.buf)
#endif

extern CLIENT *authclient1;
extern CLIENT *authclient2;
extern int refresh_authentication_handle_uid;
extern int refresh_authentication_handle_gid;
#ifdef NFSV3
extern CLIENT *authclient3;
#endif /* NFSV3 */
extern int authvers;


uid_t		auth_uid = 0;		/* Current effective user ids */
gid_t		auth_gid = 0;
GETGROUPS_T	auth_gids[NGRPS];	/* Current supplementary gids */
int		auth_gidlen = -1;
uid_t		cred_uid;
gid_t		cred_gid;
gid_t		*cred_gids;
int		cred_len;


/****************************************************************************
 * For an RPC request, look up the NFS client info along with the
 * list of directories exported to that client.
 ****************************************************************************/
nfs_client *
auth_clnt(struct svc_req *rqstp)
{
  nfs_client	*cp = NULL;
  struct in_addr addr = svc_getcaller(rqstp->rq_xprt)->sin_addr;
  
  /* Get the client and list of exports */
  if ((cp = auth_clientbyaddr(addr)) != NULL)
    return cp;
  
  /* We don't know you */
  if (trace_spoof) {
    Dprintf(L_ERROR, "Unauthorized access by NFS client %s.\n",
	    inet_ntoa(addr));
  }
  
  return (NULL);
}

/****************************************************************************
 * Check if specified client is permitted to access this path.
 * Used by mountd.c to control access to exported directories
 ****************************************************************************/
nfs_mount *
auth_path(nfs_client *cp, struct svc_req *rqstp, char *path)
{
  nfs_mount	*mp;
  
  /* Check if the specified client is permitted to access this */
  if ((mp = auth_match_mount(cp, path)) == NULL) {
    if (cp->flags != 0 || trace_spoof) {
      Dprintf(L_ERROR, "NFS client %s tried to access %s\n",
	      cp->clnt_name, path);
    }
    return NULL;
  }

  /* Check request originated on a privileged port. */
  if (!allow_non_root && mp->o.secure_port
      && !SECURE_PORT(svc_getcaller(rqstp->rq_xprt)->sin_port)) {
    Dprintf(L_ERROR,
	    "NFS request from %s originated on insecure port, %s\n",
	    cp->clnt_name,
	    "psychoanalysis suggested");
    return (NULL);
  }

  if (logging_enabled(D_AUTH)) {
    Dprintf(D_AUTH, "auth_path(%s): mount point %s, (%s%s%s%s%s)\n",
	    path, mp->path,
	    mp->o.all_squash? "all_squash " :
	    (mp->o.root_squash? "root_squash " : ""),
	    (mp->o.uidmap == map_daemon)? "uidmap " : "",
	    mp->o.secure_port? "secure " : "insecure ",
	    mp->o.link_relative? "linkrel " : "",
	    mp->o.read_only? "ro" : "rw");
  }
  
  return mp;
}

/****************************************************************************
 * Check if uid/gid of requestor can access this mount point.
 * Used by auth_fh() in nfsd.c to control access to every NFS call
 ****************************************************************************/
int auth_user(nfs_mount *mp, struct svc_req *rqstp)
{
  uid_t		cuid;
  gid_t		cgid;
  GETGROUPS_T	cgids[NGRPS];
  int		squash = mp->o.all_squash;
  int		cred_set, i, clen;

  cred_set = 0;
  if (rqstp->rq_cred.oa_flavor == AUTH_UNIX) {
    struct authunix_parms *unix_cred;
    
    unix_cred = (struct authunix_parms *) rqstp->rq_clntcred;
    cred_uid  = unix_cred->aup_uid;
    cred_gid  = unix_cred->aup_gid;
    cred_len  = unix_cred->aup_len;
    cred_gids = unix_cred->aup_gids;
    cred_set  = 1;

  }

  /* We will want to support AUTH_DES/AUTH_KRB one day,
   * but for now we treat all other authentication flavor
   * as AUTH_NULL.
   */
  if (!cred_set)
    squash = 1;
  
  if (!squash) {
    if (cred_len < 0)
      cred_len = 0;
    else if (cred_len > NGRPS)
      cred_len = NGRPS;
    /* Do the uid/gid mapping here.
     */
    /* static mapping */      
    cuid = luid(cred_uid, mp, rqstp);
    cgid = lgid(cred_gid, mp, rqstp);
    clen = cred_len;
    for (i = 0; i < cred_len; i++)
      cgids[i] = lgid(cred_gids[i], mp, rqstp);
  } else {
    /* squash this id - map it to user nobody */
    /* On systems that have 32bit uid_t in user space but
     * 16bit in the kernel, we need to truncate the
     * nobody ID (default -2).
     */
#if !defined(HAVE_BROKEN_SETFSUID)
    cuid = mp->o.nobody_uid;
    cgid = mp->o.nobody_gid;
#else
    cuid = (unsigned short) mp->o.nobody_uid;
    cgid = (unsigned short) mp->o.nobody_gid;
#endif
    /* Construct a list of one gid (nobody's gid). */
    cgids[0] = cgid;
    clen = 1;
  }


  Dprintf(D_AUTH,"auth_user: cred_uid = %i, cuid = %i\n", cred_uid, cuid);

  /* If has not seen this cuid before, use static map credentials to
     refresh uid/gid authentication hand for RPC call */
  if(refresh_authentication_handle_uid != cuid ||
     refresh_authentication_handle_gid != cgid) {
    Dprintf(D_AUTH,"auth_user: refreshing auth handle (%i,%i) - (%i,%i)\n",
	    refresh_authentication_handle_uid,
	    refresh_authentication_handle_gid,
	    cuid,cgid);
    refresh_authentication_handle_uid = cuid;
    refresh_authentication_handle_gid = cgid;
    if(authvers==1) {
      auth_destroy(authclient1->cl_auth);
      authclient1->cl_auth = create_authenticator_with_map(AUTH_UNIX,
							   cuid, cgid);
    }
    if(authvers==2) {
      auth_destroy(authclient2->cl_auth);
      authclient2->cl_auth = create_authenticator_with_map(AUTH_UNIX,
							   cuid, cgid);
    }
#ifdef NFSV3
    if(authvers==3) {
      auth_destroy(authclient3->cl_auth);
      authclient3->cl_auth = create_authenticator_with_map(AUTH_UNIX,
							   cuid, cgid);
    }
#endif /* NFSV3 */
  }
  return 1; 
}

#ifdef SOLARIS
void setfsuid(int uid) {
  return;
}
#endif /* SOLARIS */

/****************************************************************************
 * The following functions deal with setting the client's uid/gid.
 ****************************************************************************/
void
auth_override_uid(uid_t uid)
{

  return;
  /*
#if defined(HAVE_BROKEN_SETFSUID)
  uid = (unsigned short) uid;
#endif
#if defined(HAVE_SETFSUID)
  setfsuid(uid);
#else
#if defined(MAYBE_HAVE_SETFSUID)
  if (have_setfsuid)
    setfsuid(uid);
  else
#endif
    seteuid(uid);
#endif
  */
}

