/***************************************************************************
 * Copyright (c) 2002-2005, In-VIGO Team
 * Advanced Computing and Information Systems (ACIS) Laboratory
 * University of Florida
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * cache_data.c
 *
 * Functions that implement the gvfs proxy data cache.
 * The data cache is stored in the local disk.
 * Refer to cache_data.h for a description of the data cache
 * organization
 *
 * Author:	Renato J. Figueiredo
 *		Ming Zhao
 *
 * Updated: 	Jan 2004
 *		Corrected the problem that the resources allocated RPC results
 *		are not freed in several places.
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef SOLARIS
#include <sys/file.h>
#endif
#include <unistd.h>
#include <fcntl.h>
#include <math.h>

#include "nfs_prot.h"
#include "logging.h"
#include "cache_data.h"

/* 
 * Global variables
 */

unsigned long int dcache_size = MAX_CACHE_SIZE;
unsigned int dcache_assoc = MAX_CACHE_ASSOC;
unsigned int dcache_banks = MAX_CACHE_BANKS;
unsigned long int dcache_sets; /* total number of cache sets */
unsigned long int dsets_per_bank;
char *dBaseCachePath;
char *dSessionID;
char *dscratch_path;
struct stat *statbuf;
dcacheblock *dblock_scratch_RPC;
dcacheblock *dblock_scratch_disk;

/* indices used in search for a block replacement; RANDOM_PERMUTE
   elements are randomly permuted every time a new search is performed */
char *randomrepl;
#define RANDOM_PERMUTE 4

int flush_data_cache(void) {
  unsigned long int i;
  int staterr;

  Dprintf(D_DCACHE, "Flushing data cache...\n");
  for(i=0; i<dcache_banks; i++) {
    /* Not yet checking for modified blocks!!! */
    if((strlen(dBaseCachePath)+1+strlen(dSessionID)+
	1+strlen("data_cache_bank_")+16)>MAX_STR) {
      Dprintf(L_ERROR,"cache block path too long\n");
      return -1;
    }
    sprintf(dscratch_path,"%s/%s/data_cache_bank_%lu",dBaseCachePath,
	    dSessionID,i);
    staterr = stat(dscratch_path, statbuf);
    if(staterr==0) { /* file exists; remove it */
      if((unlink(dscratch_path))==-1) {
	Dprintf(L_ERROR, "Could not flush data cache bank!\n");
	return -1;
      }
    }
  }
  Dprintf(D_DCACHE, "Done flushing data cache.\n");
  return 0;
}

/*************************************************************************
 * NFS-V2-READ cache entry manipulation functions
 *************************************************************************/

/*************************************************************************
 * Copy NFS-V2 read arguments into dcacheblock data structure 
 * that is passed as argument. Returns 0 if successful, -1 if
 * error                             
 *************************************************************************/
int argument_read_v2(readargs *arg, dcacheblock *block) {

  if(arg==NULL) {
    Dprintf(D_DCACHE,"argument_read_v2: arg==NULL\n");
    return -1;
  }

  /* zero out cache block */
  bzero(block,sizeof(*block));

  /* mark block as NFS V2 read */
  block->nfsprogid = 2;
  block->nfsprocid = NFSPROC_READ;

  /* copy contents of argument to cache block */
  if((memcpy(&block->contents.read_v2.file,&arg->file,NFS_FHSIZE))==NULL) {
    Dprintf(D_DCACHE,"argument_read_v2: memcpy failed\n");
    return -1;
  }
  block->contents.read_v2.offset = arg->offset;
  block->contents.read_v2.count = arg->count;
  block->contents.read_v2.totalcount = arg->totalcount;
  return 0;
}

/*************************************************************************
 * Copy NFS-V2 read results into dcacheblock data structure 
 * that is passed as argument. Returns 0 if successful, -1 if
 * error                             
 *************************************************************************/
int response_remote_RPC_read_v2(readres *res, 
				dcacheblock *block) {
  u_int data_len;

  if(res==NULL) {
    Dprintf(D_DCACHE,"response_remote_RPC_read_v2: res==NULL\n");
    return -1;
  }

  block->contents.read_v2.status = res->status;
  data_len = res->readres_u.reply.data.data_len;
  block->contents.read_v2.data_len = data_len;

  if(data_len > 0) /*server return zero size block if the block doesn't exist*/
    if((memcpy(&block->contents.read_v2.data_val[0],
	       &res->readres_u.reply.data.data_val[0],data_len))==NULL) {
      Dprintf(D_DCACHE,"response_remote_RPC_read_v2: memcpy failed\n");
      return -1;
    }
  
  block->contents.read_v2.attributes = res->readres_u.reply.attributes;
  return 0;
}

/*************************************************************************
 * Copy NFS-V2 read cached data from dcacheblock data structure to RPC
 * response that is passed as argument. Returns 0 if successful, -1 if
 * error                             
 *************************************************************************/
int response_from_disk_cache_read_v2(dcacheblock *block, readargs *args,
				     readres *res) {
  int offset_index;
  int data_len;

  if(res==NULL) {
    Dprintf(D_DCACHE,"response_from_disk_cache_read_v2: res==NULL\n");
    return -1;
  }
  if(args==NULL) {
    Dprintf(D_DCACHE,"response_from_disk_cache_read_v2: args==NULL\n");
    return -1;
  }
  res->status = block->contents.read_v2.status;
  res->readres_u.reply.attributes = block->contents.read_v2.attributes;
  /* calculate where the offset is within the cache block */
  offset_index = args->offset % MAX_NFS_BLOCKSIZE;

  /* returned data may be more than required */
  if((block->contents.read_v2.data_len - offset_index) >= args->count)
    data_len = args->count;
  else  data_len = block->contents.read_v2.data_len - offset_index;

  res->readres_u.reply.data.data_len = data_len;
  if(data_len > 0)
    if((memcpy(&res->readres_u.reply.data.data_val[0],
  	       &block->contents.read_v2.data_val[offset_index],
  	       data_len))==NULL) {
      Dprintf(D_DCACHE,"response_from_disk_cache_read_v2: memcpy failed\n");
      return -1;
    }

  return 0;
}

/*************************************************************************
 * Original hash function; block of (offset+MAX_NFS_BLOCKSIZE) maps to   *
 * same set, but two consecutive blocks will map to different set/bank   *
 *************************************************************************/
unsigned long int hash_readargs_v2_0(nfs_fh *fhandle, u_int offset) {
  int i;
  double A=0.6180339887;
  unsigned long int B=0;

  /* accumulate file handle into B */
  for(i=0;i<NFS_FHSIZE;i++) {
    B+=(unsigned long int)(fhandle->data[i]);
  }
  B+=offset-(offset%MAX_NFS_BLOCKSIZE);

  A = A * (double)B;
  A = A - floor(A);
  A = A * dcache_sets;
  return (unsigned long int) (A) % dcache_sets;
}

/*************************************************************************
 * Attempt to improve locality for sequential file reading;              *
 * HASH_CONSECUTIVEBLOCKS__SETS block of (offset+MAX_NFS_BLOCKSIZE) map  *
 * to same set; it seems to work well for few                            *
 * file handles right now, but it is important not to have               *
 * HASH_CONSECUTIVE_SETS larger that cache's associativity, otherwise    *
 * will thrash cache                                                     *
 *************************************************************************/
unsigned long int hash_readargs_v2_1(nfs_fh *fhandle, u_int offset) {
  int i;
  double A=0.6180339887;
  unsigned long int B=0;

  /* accumulate file handle into B */
  for(i=0;i<NFS_FHSIZE;i++) {
    B+=(unsigned long int)(fhandle->data[i]);
  }
  /* effective offset is the request's offset (modulo) the product
     of the NFS block size and number of consecutive blocks that
     map to a same set */
  B+=offset-(offset%(MAX_NFS_BLOCKSIZE*HASH_CONSECUTIVE_BLOCKS_SET));

  A = A * (double)B;
  A = A - floor(A);
  A = A * dcache_sets;
  return (unsigned long int) (A) % dcache_sets;
}

/*************************************************************************
 * Another attempt to improve locality for sequential file reading;      *
 * HASH_CONSECUTIVE_BLOCKS_SETS block of (offset+MAX_NFS_BLOCKSIZE) map  *
 * to same set; HASH_CONSECUTIVE_SETS of those map to consecutive sets   *
 * file handles right now, but it is important not to have               *
 * HASH_CONSECUTIVE_SETS larger that cache's associativity, otherwise    *
 * will thrash cache                                                     *
 *************************************************************************/
unsigned long int hash_readargs_v2_2(nfs_fh *fhandle, u_int offset) {
  int i;
  double A=0.6180339887,D;
  unsigned long int B=0,C;

  /* accumulate file handle into B */
  for(i=0;i<NFS_FHSIZE;i++) {
    B+=(unsigned long int)(fhandle->data[i]);
  }
  /* effective offset is the request's offset (modulo) the product
     of the NFS block size and number of consecutive blocks that
     map to a same set */
  B+=offset-(offset%(MAX_NFS_BLOCKSIZE*HASH_CONSECUTIVE_BLOCKS_SET*
		     HASH_CONSECUTIVE_SETS));

  A = A * (double)B;
  A = A - floor(A);
  /* C is a set in the range [0,dcache_sets/HASH_CONSECUTIVE_SETS-1] */
  A = A * (dcache_sets/HASH_CONSECUTIVE_SETS);
  C = (unsigned long int) (A) % (dcache_sets/HASH_CONSECUTIVE_SETS);
  /* Expand C to range [0,dcache_sets-1], based on value of offset */
  D = floor((offset/(MAX_NFS_BLOCKSIZE*HASH_CONSECUTIVE_BLOCKS_SET)));
  C = (C * HASH_CONSECUTIVE_SETS) + ( ((unsigned long int)(D))%
				      HASH_CONSECUTIVE_SETS );
  return (C % dcache_sets);
}

/*************************************************************************
 * Hash V2 readargs (fh + offset)                                        *
 * Return set number                                                     *
 * hook to multiple possible hash functions                              *
 *************************************************************************/
unsigned long int hash_readargs_v2(nfs_fh *fhandle, u_int offset) {
  switch(HASH_FUNCTION_TO_USE) {
  case 1:
    return hash_readargs_v2_1(fhandle,offset);
    break;
  case 2:
    return hash_readargs_v2_2(fhandle,offset);
    break;
  case 0:
  default:
    return hash_readargs_v2_0(fhandle,offset);
    break;
  }
}

/*************************************************************************
 * Determine the file name that a cache block maps to                    *
 * Returns set number,                                                   *
 *  and the path to the cache bank file if successfull, -1 otherwise     *
 *************************************************************************/
long int dblock_file_path_and_set(dcacheblock *block, char *path) {
  unsigned long int hash, bank_number;

  if(block==NULL) {
    Dprintf(D_DCACHE,"dblock_file_path_and_set: block==NULL\n");
    return -1;
  }

  switch(block->nfsprogid) {
  case 2:
      hash = hash_readargs_v2(&(block->contents.read_v2.file),
			       (block->contents.read_v2.offset));
      bank_number = (unsigned long int) (hash/dsets_per_bank);
      if((strlen(dBaseCachePath)+1+strlen(dSessionID)+
	  1+strlen("data_cache_bank_")+16)>MAX_STR) {
	Dprintf(L_ERROR,"cache block path too long\n");
	return -1;
      }
      sprintf(path,"%s/%s/data_cache_bank_%lu",dBaseCachePath,dSessionID,
	      bank_number);
      Dprintf(D_DCACHE,"dcache: HASH(%lu) = %lu (bank %lu)\n",
	      block->contents.read_v2.offset,hash,bank_number);
      return (unsigned long int) hash;
  default:
    return -1;
  }
}

/*************************************************************************
 * Check whether tag of NFS arguments matches cached data on disk,       * 
 *  possibly invalidating cached data (if invalidate==1)                 *
 * block_RPC : has NFS arguments                                         *
 * block_disk: returns contents of a match on a hit                      *
 * return values:                                                        *
 * 0 : no match                                                          *
 * 1 : tag match                                                         *
 * 2 : tag match but request crosses cache block boundary                *
 * -1: error                                                             *
 *************************************************************************/
int dtag_match_and_inv(dcacheblock *block_RPC, dcacheblock *block_disk,
		       int invalidate) {
  int fd,reads,framescan;
  int file_handle_match = 0;
  int offset_match = 0;
  int overall_match = 0;
  int seekset, seekr, seekcur;
  unsigned long int set_number;
  int cachedoffset, requestoffset, staterr;

  Dprintf(D_DCACHE, "dtag_match(%d, %d)\n", block_RPC->nfsprogid,block_RPC->nfsprocid);
  if(block_RPC==NULL) {
    Dprintf(D_DCACHE,"dtag_match: block_RPC==NULL\n");
    return -1;
  }

  /* obtain path and set number */
  set_number = dblock_file_path_and_set(block_RPC, dscratch_path);
  if(set_number==-1) {
    Dprintf(D_DCACHE,"dtag_match: could not obtain set\n");
    return -1;
  }

  /* open & seek into 1st block of appropriate set in cache file */
  staterr = stat(dscratch_path, statbuf);
  if(staterr != 0) {
    Dprintf(D_DCACHE, "dtag_match: bank does not exist\n");
    return 0;
  }
  if((fd = open(dscratch_path, O_RDONLY)) == -1) {
    Dprintf(D_DCACHE,"dtag_match: could not open cache bank %s for reading\n",
	    dscratch_path);
    return -1;
  }

  /* obtain a shared lock for cache lookup */
  if ((flock(fd,LOCK_SH))==-1) {
    Dprintf(D_DCACHE,"dtag_match: could not lock cache bank %s for reading\n",
	    dscratch_path);
    return -1;
  }

  seekset = (set_number % dsets_per_bank)*dcache_assoc*sizeof(*block_RPC);
  Dprintf(D_DCACHE,
	  "dtag_match: seeking %lu -> %lu\n",set_number,seekset);
  seekr = lseek(fd,seekset,SEEK_SET);
  if(seekr != seekset) {
    Dprintf(D_DCACHE,"dtag_match: could not seek initial block %i in %s\n",
	    seekset, dscratch_path);
    flock(fd,LOCK_UN);
    close(fd);
    return -1;
  }

  /* associative lookup; traverse all sets */
  /* stops before traversing the sets if a match was found */
  for(framescan=0; (framescan<dcache_assoc && !overall_match); framescan++) {
    reads = read(fd,block_disk,sizeof(*block_disk));
    if(reads != sizeof(*block_disk)) { /* read was not right */
      Dprintf(D_DCACHE,"dtag_match: could not read scan block %i in %s\n",
	      framescan, dscratch_path);
      flock(fd,LOCK_UN);
      close(fd);
      return -1;
    }
    /* different NFS calls require diff. tag matching */
    switch(block_RPC->nfsprogid) {
    case 2:
      switch(block_RPC->nfsprocid) {
      case NFSPROC_READ:
	if (block_disk->state == ST_INV) /* invalid block; no match */
	  break;
	/* check if blocks' prog/proc ids match */
	if(!((block_RPC->nfsprogid==block_disk->nfsprogid) &&
	    (block_RPC->nfsprocid==block_disk->nfsprocid))) 
	  break;
	/* check if offsets match */
	/* a write may begin in an offset different from where a read
	   was issued; must invalidate in this case. Hence, must check
	   offset % BLOCK_SIZE. */
	cachedoffset = block_disk->contents.read_v2.offset;
	requestoffset = block_RPC->contents.read_v2.offset;
	offset_match=(((cachedoffset-(cachedoffset%MAX_NFS_BLOCKSIZE))==
		       (requestoffset-(requestoffset%MAX_NFS_BLOCKSIZE))));
	if(!offset_match) break;
	/* check if file handles match */
	file_handle_match = memcmp(&block_RPC->contents.read_v2.file,
				   &block_disk->contents.read_v2.file,
				   NFS_FHSIZE);
        if(file_handle_match != 0) break;
	overall_match = 1;

	/* invalidate block in disk if required */
	if(invalidate) {
	  /* keep current seek point in seekcur; will need it after
	     fd is reopened */
	  seekcur = lseek(fd, 0, SEEK_CUR);
	  /* release shared lock  */
	  if ((flock(fd,LOCK_UN))==-1) {
	    Dprintf(D_DCACHE,
		    "dtag_match: could not unlock cache bank %s for reading\n",
		    dscratch_path);
	    return -1;
	  }
	  /* reopen bank file in read/write mode  for invalidation */
	  close(fd);
	  if((fd = open(dscratch_path, O_RDWR)) == -1) {
	    Dprintf(D_DCACHE,"dtag_match: could not open cache bank %s R/W\n",
		    dscratch_path);
	    return -1;
	  }
	  /* obtain exclusive lock for invalidation */
	  if ((flock(fd,LOCK_EX))==-1) {
	    Dprintf(D_DCACHE,
		    "dtag_match: could not lock cache bank %s for writing\n",
		    dscratch_path);
	    return -1;
	  }
	  /* seek back one block from seekcur */
	  seekset = seekcur-sizeof(*block_RPC); /* point back one block */
	  seekr = lseek(fd, seekset, SEEK_SET);
	  if(seekr!=seekset) {
	    Dprintf(D_DCACHE,"dtag_match: could not seek %i in file bank %s\n",
		    seekset, dscratch_path);
	    flock(fd,LOCK_UN);
	    close(fd);
	    return -1;
	  }
	  /* write an invalid block (all zeroes) */
	  bzero(block_disk,sizeof(*block_disk));
	  if((write(fd,block_disk,sizeof(*block_disk)))!=sizeof(*block_disk)) {
	    Dprintf(D_DCACHE,"dtag_match: could not invalidate block in %s\n",
		    dscratch_path);
	    flock(fd,LOCK_UN);
	    close(fd);
	    return -1;
	  }	  
	}/* endif invalidate */
        break;
      } /*end switch(block_RPC->nfsprocid)*/
      break;
    } /*end switch(block_RPC->nfsprogid)*/
  }/* end for framescan */
  /* release shared lock  */
  if ((flock(fd,LOCK_UN))==-1) {
    Dprintf(D_DCACHE,
	    "dtag_match: could not unlock cache bank %s\n", dscratch_path);
    return -1;
  }
  close(fd);

  Dprintf(D_DCACHE,"dtag_match(%d): offset(%lu) returns %i (frame %i)\n",
	    block_RPC->nfsprocid,
	    block_RPC->contents.read_v2.offset,overall_match,framescan-1);
  return overall_match;
}

/*************************************************************************
 * Finds if there is a frame with specified state for a cache block      *
 * creates new file for cache bank if necessary                          *
 * return values:                                                        *
 * 0 =< x < cache_assoc        : free frame found in x                   *
 * cache_assoc:                : no free frames found                    *
 * -1: error                                                             *
 *************************************************************************/
int dfind_replace_frame(unsigned long int set_number, dcacheblock *block_disk, 
			int state) {
  int fd,reads,framescan;
  int seekset, seekr;
  int idx1,idx2,i,ctmp;

  /* open & seek into 1st block of appropriate set in cache file */
  if((fd = open(dscratch_path, O_RDONLY)) == -1) {
    Dprintf(D_DCACHE,
	    "dfind_replace_frame: could not open cache bank %s for reading\n",
	    dscratch_path);
    return -1;
  }
  /* obtain a shared lock for cache lookup */
  if ((flock(fd,LOCK_SH))==-1) {
    Dprintf(D_DCACHE,
	    "dfind_replace_frame: could not lock cache bank %s for reading\n",
	    dscratch_path);
    return -1;
  }

  seekset = (set_number % dsets_per_bank)*dcache_assoc*sizeof(dcacheblock);
  Dprintf(D_DCACHE,
	  "dfind_replace_frame: seeking %lu -> %lu\n",set_number,seekset);

  seekr = lseek(fd,seekset,SEEK_SET);
  if(seekr != seekset) {
    Dprintf(D_DCACHE,
	    "dfind_replace_frame: could not seek initial block %i in %s\n",
	    seekset, dscratch_path);
    flock(fd,LOCK_UN);
    close(fd);
    return -1;
  }

  /* randomly permute elements of randomrepl */
  for(i=0;i<RANDOM_PERMUTE;i++) {
    idx1 = random()%dcache_assoc;
    idx2 = random()%dcache_assoc;
    ctmp = randomrepl[idx1];
    randomrepl[idx1] = randomrepl[idx2];
    randomrepl[idx2] = ctmp;
  }

  /* associative lookup; traverse all sets */
  for(i=0; i<dcache_assoc; i++) {

    if(state==ST_INV) {
      /* linear scan; no need to seek */
      framescan = i;
    } else {
      /* random scan; need to seek appropriate block */
      framescan = randomrepl[i];
      seekr = lseek(fd,(seekset+framescan*sizeof(dcacheblock)),SEEK_SET);
      if(seekr != seekset+framescan*sizeof(dcacheblock)) {
	Dprintf(D_DCACHE,
		"dfind_replace_frame: could not seek random block %i in %s\n",
		framescan, dscratch_path);
	flock(fd,LOCK_UN);
	close(fd);
	return -1;
      }
    }

    reads = read(fd,block_disk,sizeof(*block_disk));
    if(reads != sizeof(*block_disk)) { /* read was not right */
      Dprintf(D_DCACHE,
	      "find_replace_frame: could not read scan block %i in %s\n",
	      framescan, dscratch_path);
      flock(fd,LOCK_UN);
      close(fd);
      return -1;
    }
    if(block_disk->state == state) {
      /* release shared lock  */
      if ((flock(fd,LOCK_UN))==-1) {
	Dprintf(D_DCACHE,
		"dfind_replace_frame: could not unlock cache bank %s\n",
		dscratch_path);
	return -1;
      }
      close(fd);
      Dprintf(D_DCACHE,
	      "dfind_replace_frame: found %i (state %d)\n", 
	      framescan, block_disk->state);
      return framescan;
    }
  }

  /* release shared lock  */
  if ((flock(fd,LOCK_UN))==-1) {
    Dprintf(D_DCACHE,
	    "dfind_replace_frame: could not unlock cache bank %s\n",
	    dscratch_path);
    return -1;
  }
  close(fd);
  return dcache_assoc;
}

/*************************************************************************
 * Insert result from RPC call in disk cache                             *
 * 0 : inserted                                                          *
 *-1 : error                                                             *
 *************************************************************************/
int dinsert_in_cache(dcacheblock *block_RPC, dcacheblock *block_disk) {
  int freeset,replaceset,foundset;
  unsigned long int set_number;
  int staterr, fd;
  int written, seekset, seekr, seekcur;
  char c = '\0';

  if(block_RPC==NULL)
    return -1;

  /* determine if cache bank exists; if not, create it */
  set_number = dblock_file_path_and_set(block_RPC, dscratch_path);
  staterr = stat(dscratch_path, statbuf);

  if(staterr!=0) { /* create new bank */
    if((fd = creat(dscratch_path, S_IRUSR | S_IWUSR |
		   S_IRGRP | S_IWGRP ))==-1) {
      Dprintf(D_DCACHE,"dinsert_in_cache: could not create %s for R/W\n",
	      dscratch_path);
      return -1;
    }
    lseek(fd, dsets_per_bank*dcache_assoc*sizeof(*block_disk)-1, SEEK_SET);
    if(write(fd, &c, 1) != 1) {
      Dprintf(D_DCACHE,"dinsert_in_cache: could not write to %s\n",
	      dscratch_path);
      close(fd);
      return -1;
    }
    close(fd);
  } else { /* bank file exists; check permissions */
    if(!((statbuf->st_mode & S_IRUSR) &&
	 (statbuf->st_mode & S_IWUSR))) {
      Dprintf(D_DCACHE,"dinsert_in_cache: wrong permissions in %s\n",
	      dscratch_path);
      return -1;
    }
  }

  /* first try, look for empty block */
  if((freeset = dfind_replace_frame(set_number, block_disk, ST_INV))==-1) {
    Dprintf(D_DCACHE,"dinsert_in_cache: error finding replace frame ST_INV\n");
    return -1;
  }
  if(freeset == dcache_assoc) { /* second try, look for clean block */
    Dprintf(D_DCACHE, "No empty blocks\n");
    if((replaceset = dfind_replace_frame(set_number, block_disk, 
					 ST_RO_CLEAN))==-1) {
      Dprintf(D_DCACHE,
	      "dinsert_in_cache: error finding replace frame ST_RO_CLEAN\n");
      return -1;
    }
    if(replaceset == dcache_assoc) { /* mission impossible */	
      Dprintf(D_DCACHE,
	      "dinsert_in_cache: could not find block to evict\n");
      return -1;
    }
    foundset = replaceset;
    Dprintf(D_DCACHE,"dinsert_in_cache: offset(%lu) replaces block %i\n",
	    block_RPC->contents.read_v2.offset,replaceset);
  } else {
    foundset = freeset;
    Dprintf(D_DCACHE,"dinsert_in_cache: offset(%lu) found empty frame %i\n",
	    block_RPC->contents.read_v2.offset,freeset);
  }

  /* open cache bank */
  if((fd = open(dscratch_path, O_RDWR))==-1) {
    Dprintf(D_DCACHE, "dinsert_in_cache: could not open file bank %s\n",
	    dscratch_path);
    return -1;
  }

  /* obtain exclusive lock for insertion */
  if ((flock(fd,LOCK_EX))==-1) {
    Dprintf(D_DCACHE,
	    "dinsert_in_cache: could not lock cache bank %s for writing\n",
	    dscratch_path);
    return -1;
  }

  /* seek into appropriate insert position */
  /* first, seek into first block of set */
  seekset = (set_number % dsets_per_bank)*dcache_assoc*sizeof(*block_RPC);
  seekr = lseek(fd,seekset,SEEK_SET);
  if(seekr != seekset) {
    Dprintf(D_DCACHE,
	    "dinsert in cache: could not seek initial block %i in %s\n",
	    seekset, dscratch_path);
    flock(fd,LOCK_UN);
    close(fd);
    return -1;
  }

  /* then, seek within set for replaced frame */
  seekcur = seekr; /* remember current seek point */
  seekset = foundset*sizeof(*block_RPC); /* determine offset within set */
  seekr = lseek(fd, seekset, SEEK_CUR); /* offset seek */
  if((seekcur+seekset) != seekr) {
    Dprintf(D_DCACHE, "dinsert_in_cache: could not seek %i in file bank %s\n",
	    seekset, dscratch_path);
    flock(fd,LOCK_UN);
    close(fd);
    return -1;
  }

  /* determine cache block state (RO, RW) */
  switch(block_RPC->nfsprogid) {
  case 2:
    switch(block_RPC->nfsprocid) {
    case NFSPROC_READ:
      block_RPC->state = ST_RO_CLEAN;
      break;
    default:
      block_RPC->state = ST_INV;
    }
    break;
  case 3:
  default:
    block_RPC->state = ST_INV;
  }

  /* write block into cache */
  written = write(fd, block_RPC, sizeof(*block_RPC));
  if(written != sizeof(*block_RPC)) {
    Dprintf(L_ERROR,"Could not write into cache!\n");
    flock(fd,LOCK_UN);
    close(fd);
    return -1;
  }
  /* release exclusive lock  */
  if ((flock(fd,LOCK_UN))==-1) {
    Dprintf(D_DCACHE,
	    "dinsert_in_cache: could not unlock cache bank %s for writing\n",
	    dscratch_path);
    return -1;
  }
  close(fd);
  return 0;
}

/*************************************************************************
 * Initialize global variables                                           *
 * path    : base path to file cache directory                           *
 * return values:                                                        *
 * 0 : successfull                                                       *
 * -1 : unsuccessful                                                     *
 *************************************************************************/
int init_data_cache_global(char *basepath, char *sessionid) {
  int staterr, mkdirerr, i;

  if (!(dscratch_path = (char *)malloc(sizeof(*dscratch_path)*MAX_STR)))
    return -1;

  if (!(dSessionID = (char *)malloc(sizeof(*dSessionID)*MAX_STR)))
    return -1;
  if(strlen(sessionid)>(sizeof(*dSessionID)*MAX_STR))
    return -1;
  strcpy(dSessionID,sessionid);

  if (!(dBaseCachePath = (char *)malloc(sizeof(*dBaseCachePath)*MAX_STR)))
    return -1;
  if(strlen(basepath)>(sizeof(*dBaseCachePath)*MAX_STR))
    return -1;
  strcpy(dBaseCachePath,basepath);

  if (!(statbuf = (struct stat *) malloc (sizeof(*statbuf))))
    return -1;

  if (!(dblock_scratch_RPC=(dcacheblock *)
	malloc(sizeof(*dblock_scratch_RPC))))
    return -1;

  if (!(dblock_scratch_disk=(dcacheblock *)
	malloc(sizeof(*dblock_scratch_disk))))
    return -1;

  if ((randomrepl=malloc(sizeof(*randomrepl)*dcache_assoc)) == NULL)
    return -1;
  for(i=0;i<dcache_assoc;i++)
    randomrepl[i]=i;

  /* check the existence of base path and session ID directories */
  staterr = stat(dBaseCachePath, statbuf);
  if(!((staterr==0) && (statbuf->st_mode & S_IRUSR) &&
       (statbuf->st_mode & S_IWUSR))) {
    Dprintf(L_ERROR, "dBaseCachePath directory %s does not exist\n",
	    dBaseCachePath);
    return -1;
  }
  if((strlen(dBaseCachePath)+1+strlen(dSessionID)+
      1+strlen("data_cache_bank_")+16)>MAX_STR) {
    Dprintf(L_ERROR,"cache block path too long\n");
    return -1;
  }
  sprintf(dscratch_path,"%s/%s",dBaseCachePath,dSessionID);
  staterr = stat(dscratch_path, statbuf);  
  if(!((staterr==0) && (statbuf->st_mode & S_IRUSR) &&
       (statbuf->st_mode & S_IWUSR))) {
    mkdirerr = mkdir(dscratch_path,S_IRWXU | S_IRWXG);
    if(mkdirerr==-1) {
      Dprintf(L_ERROR, "Could not create dSessionID directory %s\n",
	      dscratch_path);
      return -1;
    }
  }
  return 0;
}

void dump_data_cache_parameters(FILE *fp) {
  fprintf(fp,"Data cache parameters:\n");
  fprintf(fp,"\tCache size: %lu blocks\n",dcache_size);
  fprintf(fp,"\tCache assoc: %du\n",dcache_assoc);
  fprintf(fp,"\t# cache banks: %du\n",dcache_banks);
  fprintf(fp,"\t# cache sets: %lu\n",dcache_sets);
  fprintf(fp,"\t# sets per bank: %lu\n",dsets_per_bank);
  fprintf(fp,"\tSize of cache block: %i\n",(int)sizeof(*dblock_scratch_RPC));
  fprintf(fp,"\tSize of cache bank file: %lu\n",
	  dsets_per_bank*dcache_assoc*sizeof(*dblock_scratch_RPC));
}

/*************************************************************************
 * Initialize cache parameters                                           *
 * size    : cache size (# entries, pwr of 2)                            *
 * assoc   : cache associativity (pwr of 2)                              *
 * return values:                                                        *
 * 0 : successfull                                                       *
 *-1 : unsuccessful                                                      *
 *************************************************************************/
int init_data_cache(unsigned long int size, int assoc, int banks,
		    char *basepath, char *sessionid) {

  unsigned long int localsize, localassoc, localbanks;

  if(basepath==NULL || sessionid==NULL)
    return -1;

  if(size==0 || size > MAX_CACHE_SIZE)
    return -1;
  if (assoc==0 || assoc > MAX_CACHE_ASSOC)
    return -1;

  if(init_data_cache_global(basepath,sessionid) == -1)
    return -1;

  localsize=size;
  while (localsize>1) {
    if(localsize%2) {
      Dprintf(L_ERROR,"Dcache size not a power of two\n");
      return -1;
    }
    localsize>>=1;
  }

  localassoc=assoc;
  while (localassoc>1) {
    if(localassoc%2) {
      Dprintf(L_ERROR,"Dcache associativity not a power of two\n");
      return -1;
    }
    localassoc>>=1;
  }

  localbanks=banks;
  while (localbanks>1) {
    if(localbanks%2) {
      Dprintf(L_ERROR,"Dcache banks not a power of two\n");
      return -1;
    }
    localbanks>>=1;
  }

  if((banks*assoc)>size) {
      Dprintf(L_ERROR,"Dcache size < banks * associativity \n");
    return -1;
  }

  dcache_size = size;
  dcache_assoc = assoc;
  dcache_banks = banks;
  dcache_sets = dcache_size/dcache_assoc;
  dsets_per_bank = dcache_size/(dcache_assoc*dcache_banks);
  Dprintf(D_DCACHE, "data cache init successful\n"
                    "size: %d, assoc: %d, banks: %d "
		    "sets: %d, sets per bank: %d\n",
		    dcache_size, dcache_assoc, dcache_banks,
		    dcache_sets, dsets_per_bank);

  return 0;
}

/*************************************************************************
 * lookup block in cache                                                 *
 * prog_index, proc_index: type of RPC call                              *
 * argument = (void *)rpcargs from incoming RPC call                     *
 * return values:                                                        *
 * 0 : miss                                                              *
 * 1 : hit                                                               *
 * -1: error                                                             *
 * side-effects:                                                         *
 *  block_scratch_RPC contains copy of diropargs                         *
 *  block_scratch_disk contains copy of results in case of a hit         *
 *  res contains RPC result looked up from cache                         *
 *************************************************************************/
int dcache_lookup(int prog_index, int proc_index, void *rpcargs, void *res) {
  int match = 0;
  readargs args;

  switch(prog_index) {
  case 2:
    switch(proc_index) {
    case NFSPROC_READ:
      /* create cache_scratch_RPC from rpcargs */
      args = *(readargs *)rpcargs;

      if((argument_read_v2(&args, dblock_scratch_RPC))==-1) {
	Dprintf(D_DCACHE,"dcache_lookup: argument copy failed\n");
	return -1;
      }
      if((match=dtag_match_and_inv(dblock_scratch_RPC,
				   dblock_scratch_disk, 0))==-1) {
	Dprintf(D_DCACHE,"dcache_lookup: tag match failed\n");
	return -1;
      }
      if (match == 1) { /* hit; copy data */
	if((response_from_disk_cache_read_v2(dblock_scratch_disk,
					     (readargs *)rpcargs,
					     (readres *)res))==-1) {
	  Dprintf(D_DCACHE,"dcache_lookup: response copy failed\n");
	  return -1;
	}
      }
      break;
    } /* switch (proc_index) */
    break;
  } /* switch(prog_index) */
  return(match);
}

/*************************************************************************
 * insert block in cache                                                 *
 * prog_index, proc_index: type of RPC call                              *
 * argument = (void *)rpcargs from incoming RPC call                     *
 * return values:                                                        *
 * 0 : inserted                                                          *
 *-1 : did not insert                                                    *
 * side-effects:                                                         *
 *  block_scratch_RPC contains copy of results                           *
 *  block_scratch_disk contains copy of results in case of a hit         *
 *************************************************************************/
int dcache_insert(void *argp, int prog_index, int proc_index, void *res) {
  readargs *readarg;
  readres *rdres;
  int base_offset,req_offset,req_count;
  int total_count;

  switch(prog_index) {
  case 2:
    rdres = (readres *)res;
    readarg = (readargs *)argp;
    /* calculate initial offset to fetch from */
    base_offset = (readarg->offset)-((readarg->offset)%MAX_NFS_BLOCKSIZE);
    /* calculate offset for building result structure */
    req_offset = (readarg->offset)%MAX_NFS_BLOCKSIZE;
    req_count = readarg->count;
    total_count = rdres->readres_u.reply.data.data_len;
    Dprintf(D_DCACHE, "dcache_insert: offset %d, count %d\n",
    	base_offset, total_count);
    
    switch(proc_index) {      
    case NFSPROC_READ:            
      /* add res. to cache_scratch_RPC*/
      if((response_remote_RPC_read_v2(rdres,
				      dblock_scratch_RPC))==-1) {
	Dprintf(D_DCACHE,"dcache_insert: RPC response copy failed\n");
	return -1;
      }
      /* prepare dblock_scratch_RPC for insertion. The request that
       actually went out to the server was for base_offset + MAX_BLOCK;
       and has returned a block of size total_count<=MAX_BLOCK */
      dblock_scratch_RPC->contents.read_v2.offset = base_offset;
      dblock_scratch_RPC->contents.read_v2.file = readarg->file;
      if((dinsert_in_cache(dblock_scratch_RPC, dblock_scratch_disk))==-1) {
	Dprintf(D_DCACHE,"dcache_insert: insertion failed\n");
	return -1;
      }
      /* prepare result structure; we need to overwrite *res so that its
         first data matches the requested data, and the data length
         matches the requested count
         two situations can arise: req_count + offset is <= than the count
	 returned by the server (total_count), in which case all requested
         bytes have been returned and we simply return req_count bytes. 
         Or, the total bytes returned is not sufficient to cover
	 req_offset -> req_offset_req_count.
         In this case, the total number of bytes returned by the server
         is total_count, but only (total_count-req_offset) are
         visible to the client */
      if((req_count+req_offset) <= total_count)
	rdres->readres_u.reply.data.data_len = req_count;
      else if((total_count-req_offset) > 0)
	rdres->readres_u.reply.data.data_len = total_count-req_offset;
      else rdres->readres_u.reply.data.data_len = 0;

      if(rdres->readres_u.reply.data.data_len > 0)
        if((memcpy(&rdres->readres_u.reply.data.data_val[0],
	       &dblock_scratch_RPC->contents.read_v2.data_val[req_offset],
	       rdres->readres_u.reply.data.data_len))==NULL) {
	  Dprintf(D_DCACHE,"dcache_insert: memcpy error\n");
	  return -1;
      }
      return 0;
    default:
      break;
    } /* switch(proc_index) */
    break;
  } /* switch(prog_index) */
  Dprintf(D_DCACHE,"dcache_insert: unknown RPC call\n");
  return -1;
}

/*************************************************************************
 * Lookup entry for readarg *arg in data cache.
 * if found, invalidate it
 * return values:                                                        *
 * 0 : no invalidation needed                                            *
 * 1 : successful invalidation                                           *
 * -1: error                                                             *
 *************************************************************************/
int invalidate_readarg_v2(readargs *arg) {
  int match1 = 0, match2 = 0;
  readargs args;
  int crossed_boundary = 0;

  if(arg==NULL) {
    Dprintf(D_DCACHE,"invalidate_readarg_v2: arg==NULL\n");
    return -1;
  }

  args = *arg; 
  Dprintf(D_DCACHE, "invalidate_readarg_v2: offset %d, count %d\n",
  			args.offset, args.count);			
  if(((args.offset % MAX_NFS_BLOCKSIZE) + args.count) > 
      	    MAX_NFS_BLOCKSIZE) {
    Dprintf(D_DCACHE, "invalidate_readarg: crossed block boundary\n");
    crossed_boundary = 1;
    args.offset = args.offset - (args.offset % MAX_NFS_BLOCKSIZE);
    args.count = MAX_NFS_BLOCKSIZE;
    args.totalcount = MAX_NFS_BLOCKSIZE;
  }  
  /* copy diropargs to block_scratch_RPC */
  if((argument_read_v2(&args, dblock_scratch_RPC))==-1) {
    Dprintf(D_DCACHE,"invalidate_readarg_v2: argument copy failed\n");
    return -1;
  }
  Dprintf(D_DCACHE, "invalidate_readarg_v2: invalidate offset %d\n", args.offset);
  if((match1=dtag_match_and_inv(dblock_scratch_RPC,
			      dblock_scratch_disk, 1))==-1) {
    Dprintf(D_DCACHE,"invalidate_readarg_v2: tag invalidate failed\n");
    return -1;
  }

  if(crossed_boundary) { /* invalidate second block if this access crosses
		            block boundary*/
    args.offset = arg->offset + arg->count;
    args.offset = args.offset - (args.offset % MAX_NFS_BLOCKSIZE);
    if((argument_read_v2(&args, dblock_scratch_RPC))==-1) {
      Dprintf(D_DCACHE,"invalidate_readarg_v2: argument copy error\n");
      return -1;
    }  
    Dprintf(D_DCACHE, "invalidate: offset %d\n", args.offset);  
    if((match2=dtag_match_and_inv(dblock_scratch_RPC,
				 dblock_scratch_disk, 1))==-1) {
      Dprintf(D_DCACHE,"invalidate_readarg_v2: tag match error\n");
      return -1;
    }
  }
  return (match1+match2);
}

/*************************************************************************
 * check if need to invalidate block in cache                            *
 * prog_index, proc_index: type of RPC call                              *
 * argument = (void *)rpcargs from incoming RPC call                     *
 * return values:                                                        *
 * 0 : no invalidation needed                                            *
 * 1 : successful invalidation                                           *
 * -1: error                                                             *
 * side-effects:                                                         *
 * block_scratch_RPC contains copy of rpcargs                            *
 *************************************************************************/
int dcache_inval(int prog_index, int proc_index, void *rpcargs) {
  writeargs *wargs;
  readargs  invalidate_arg, *invalidate_arg_ptr;
  int invres=0;

  switch(prog_index) {
  case 2: /* NFS-V2 */
    switch(proc_index) {
    case NFSPROC_READ:
      invalidate_arg_ptr = (readargs *)rpcargs;
      break;
    case NFSPROC_WRITE:
      wargs = (writeargs *)rpcargs;
      invalidate_arg_ptr = &invalidate_arg;
      invalidate_arg_ptr->file = wargs->file;
      invalidate_arg_ptr->offset = wargs->offset;
      invalidate_arg_ptr->count = wargs->data.data_len;
      invalidate_arg_ptr->totalcount = wargs->data.data_len;
      break;
    default:
      /* no invalidations necessary for other calls */
      invalidate_arg_ptr = NULL;
    }
    /* now perform actual invalidation */
    if(invalidate_arg_ptr!=NULL) {
      invres = invalidate_readarg_v2(invalidate_arg_ptr);
      if(invres == -1) {
	Dprintf(D_DCACHE,"Error invalidating data cache block\n");
	return -1;
      }
      Dprintf(D_DCACHE,"dcache_inval: result of argument1: %i\n", invres);
      return(invres>0);
    }
    break; /* case 2 */
  case 3: /* NFS-V3 */
    break;
  default:
    return -1; /* unknown NFS version */
  }
  return -1;
}
