/*
 * mount_dispatch.c	This file contains the function dispatch table.
 *                      This table is used to select which function
 *                      is called when an incoming mount call is received.
 *                      The actual function bodies are defined in mountd.c
 *                      and are named mountroc_FUNCTNAME_i_svc (i=1, 2, 3 for
 *                      mount protocol versions 1, 2, 3)
 *
 * Author:	Renato J. Figueiredo
 * 
 * Updated:	Corrected the problem that error code is not returned to client.
 *		Ming Zhao, June 2005
 *
 * Based on original user-level NFSD code by:
 *
 * Authors:	Donald J. Becker, <becker@super.org>
 *		Rick Sladkey, <jrs@world.std.com>
 *		Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *		Olaf Kirch, <okir@monad.swb.de>
 *
 *		This software may be used for any purpose provided
 *		the above copyright notice is retained.  It is supplied
 *		as is, with no warranty expressed or implied.
 */

#include "mountd.h"
#include "rpcmisc.h"

#ifdef SOLARIS
#define PORTMAP
#include <rpc/rpc.h>
#include <rpc/svc_soc.h>
#endif /* SOLARIS */

#ifndef svc_getcaller
#define svc_getcaller(x) ((struct sockaddr_in *) (x)->xp_rtaddr.buf)
#endif

/*
 * These are the global variables that hold all argument and result data.
 */
union argument_types 	argument;
union result_types	result;

/*
 * MOUNT versions supported by this implementation
 */
#ifdef NFSV3
#define MAXVERS         3
#else
#define	MAXVERS		2
#endif /* NFSV3 */

extern int mountvers;

/*
 * This is a dispatch table to simplify error checking,
 * and supply return attributes for NFS functions.
 */

#ifdef __STDC__
#define CONCAT(a,b)		a##b
#define CONCAT3(a,b,c)		a##b##c
#define CONCAT5(a,b,c,d,e)	a##b##c##d##e
#define STRING(a)		#a
#else
#define CONCAT(a,b)		a/**/b
#define CONCAT3(a,b,c)		a/**/b/**/c
#define CONCAT5(a,b,c,d,e)	a/**/b/**/c/**/d/**/e
#define STRING(a)		"a"
#endif

#define table_ent(vers, res_type, arg_type, funct) {		\
	sizeof(res_type), sizeof(arg_type),			\
	(xdrproc_t) CONCAT(xdr_,res_type),			\
	(xdrproc_t) CONCAT(xdr_,arg_type),			\
	(void *(*)()) CONCAT5(mountproc_,funct,_,vers,_svc),	\
	STRING(funct), CONCAT(pr_,arg_type)			\
}

#ifdef NFSV3
#define table_ent3(vers, res_type, arg_type, funct) {		\
	sizeof(res_type), sizeof(arg_type),			\
	(xdrproc_t) CONCAT(xdr_,res_type),			\
	(xdrproc_t) CONCAT(xdr_,arg_type),			\
	(void *(*)()) CONCAT5( mountproc3_,funct,_,vers,_svc),	\
	STRING(funct), CONCAT(pr_,arg_type)			\
}
#endif /* NFSV3 */

#define nil	void
#define xdr_nil	xdr_void
#define pr_nil	pr_void
#define pr_char	pr_void

extern CLIENT *mountdclient1;
extern CLIENT *mountdclient2;
#ifdef NFSV3
extern CLIENT *mountdclient3;
#endif /* NFSV3 */

struct dispatch_entry {
    int		res_size, arg_size;	/* sizeof the res/arg structs	*/
    xdrproc_t	xdr_result;
    xdrproc_t	xdr_argument;
    void	*(*funct)();		/* function handler		*/
    char	*name;			/* name of function		*/
    char	*(*log_print)();	/* ptr to debug handler		*/
};

static char *	pr_void(void);
static char *	pr_dirpath(dirpath *argp);

static struct dispatch_entry mount_1_table[] = {
	table_ent(1,nil,nil,null),			/* NULL */
	table_ent(1,fhstatus,dirpath,mnt),		/* MNT */
	table_ent(1,mountlist,void,dump),		/* DUMP */
	table_ent(1,void,dirpath,umnt),			/* UMNT */
	table_ent(1,void,void,umntall),			/* UMNTALL */
	table_ent(1,exports,void,export),		/* EXPORT */
	table_ent(1,exports,void,exportall),		/* EXPORTALL */
};

static struct dispatch_entry mount_2_table[] = {
	table_ent(2,nil,nil,null),			/* NULL */
	table_ent(2,fhstatus,dirpath,mnt),		/* MNT */
	table_ent(2,mountlist,void,dump),		/* DUMP */
	table_ent(2,void,dirpath,umnt),			/* UMNT */
	table_ent(2,void,void,umntall),			/* UMNTALL */
	table_ent(2,exports,void,export),		/* EXPORT */
	table_ent(2,exports,void,exportall),		/* EXPORTALL */
	table_ent(2,ppathcnf,dirpath,pathconf),		/* PATHCONF */
};

#ifdef NFSV3
static struct dispatch_entry mount_3_table[] = {
	table_ent3(3,nil,nil,null),			/* NULL */
	table_ent3(3,mountres3,dirpath,mnt),		/* MNT */
	table_ent3(3,mountlist,void,dump),		/* DUMP */
	table_ent3(3,void,dirpath,umnt),		/* UMNT */
	table_ent3(3,void,void,umntall),		/* UMNTALL */
	table_ent3(3,exports,void,export),		/* EXPORT */
};
#endif /* NFSV3 */

static struct dispatch_entry * dtable[MAXVERS] = {
	mount_1_table,
	mount_2_table,
#ifdef NFSV3
	mount_3_table,
#endif /* NFSV3 */
};

static unsigned int		dtnrprocs[MAXVERS] = {
	sizeof(mount_1_table) / sizeof(mount_1_table[0]),
	sizeof(mount_2_table) / sizeof(mount_2_table[0]),
#ifdef NFSV3
	sizeof(mount_3_table) / sizeof(mount_3_table[0]),
#endif /* NFSV3 */
};

/*
 * The main dispatch routine.
 */
void
mount_dispatch(struct svc_req *rqstp, SVCXPRT *transp)
{
  unsigned int		proc_index, vers_index;
  unsigned int            prog_index;
  struct dispatch_entry	*dtbl, *dent;
  union result_types	*resp;
  struct sockaddr_in	*sin;
  
  sin = (struct sockaddr_in *) svc_getcaller(transp);
  
  proc_index = rqstp->rq_proc;
  vers_index = rqstp->rq_vers - 1;
  prog_index = rqstp->rq_vers;
  _rpcsvcdirty = 1;
  
  if (prog_index != mountvers) {
    svcerr_progvers(transp, mountvers, mountvers);
    goto done;
  }

  if (proc_index >= dtnrprocs[vers_index]) {
    svcerr_noproc(transp);
    goto done;
  }
  dtbl = dtable[vers_index];
  dent = &dtbl[proc_index];
  
  memset(&argument, 0, dent->arg_size);
  if (!svc_getargs(transp, (xdrproc_t) dent->xdr_argument, (caddr_t) &argument)) {
    svcerr_decode(transp);
    goto done;
  }
  /* Clear the result structure. */
  memset(&result, 0, dent->res_size);
  
  /* Log the call. */
  if (logging_enabled(D_CALL)) {
    Dprintf(D_CALL,"prog_index=%i, proc_index=%i\n",prog_index,proc_index);
    log_call(rqstp, dent->name, dent->log_print(&argument));
  }

  /* Do the function call itself. */
  resp = (*dent->funct) (&argument, rqstp);
  Dprintf(D_CALL, "result: %i\n",result.fstatus);
  
  /*svc_sendreply(transp, dent->xdr_result, (caddr_t) resp);*/
  if (!svc_sendreply(transp, dent->xdr_result, (caddr_t) resp))
    svcerr_systemerr(transp);

  switch(prog_index) {
  case 1 :
    if (!clnt_freeres(mountdclient1,
		      (xdrproc_t) dent->xdr_result,
		      (caddr_t)&result)) {
      Dprintf(L_ERROR, "unable to free RPC arguments, exiting\n");
      exit(-1);
    }
    break;
  case 2 :
    if (!clnt_freeres(mountdclient2,
		      (xdrproc_t) dent->xdr_result,
		      (caddr_t)&result)) {
      Dprintf(L_ERROR, "unable to free RPC arguments, exiting\n");
      exit(-1);
    }
    break;
#ifdef NFSV3
  case 3 :
    if (!clnt_freeres(mountdclient3,
		      (xdrproc_t) dent->xdr_result,
		      (caddr_t)&result)) {
      Dprintf(L_ERROR, "unable to free RPC arguments, exiting\n");
      exit(-1);
    }
    break;
#endif /* NFSV3 */
  }

  if (!svc_freeargs(transp, (xdrproc_t) dent->xdr_argument, (caddr_t) &argument)) {
    Dprintf(L_ERROR, "unable to free RPC arguments, exiting\n");
    exit(-1);
  }
  
 done:
  _rpcsvcdirty = 0;
  if (need_reinit) {
    reinitialize(0);
  }
}

/*
 * Functions for debugging output.
 */
static char *pr_void()
{
  return ("");
}

static char *pr_dirpath(argp)
dirpath *argp;
{
  return (*argp);
}

