/*
 * nfs_dispatch.c	This file contains the function dispatch table.
 *                      This table is used to select which function
 *                      is called when an incoming NFS/RPC call is received.
 *                      The actual function bodies are defined in nfsd.c
 *                      and are named nfsd_nfsproc_FUNCTNAME_i (i==2 for
 *                      NFS-V2 and 3 for NFS-V3)
 * Author: Renato J. Figueiredo
 * 
 * Based on original user-level NFSD code by:
 *
 * Authors:	Donald J. Becker, <becker@super.org>
 *		Rick Sladkey, <jrs@world.std.com>
 *		Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *		Olaf Kirch, <okir@monad.swb.de>
 *
 *		This software may be used for any purpose provided
 *		the above copyright notice is retained.  It is supplied
 *		as is, with no warranty expressed or implied.
 */

#include "nfsd.h"
#include "nfs_prot.h"
#include "rpcmisc.h"
#include "time.h"

/*
 * These are the global variables that hold all argument and result data.
 */
union argument_types 	argument;
union result_types	result;

/* Global variables for V2/V3 NFS clients */
extern CLIENT *nfsdclient2;
#ifdef NFSV3
extern CLIENT *nfsdclient3;
#endif /* NFSV3 */
extern int nfsvers;

/*
 * The time at which we received the request.
 * Useful for various book-keeping things
 */
time_t			nfs_dispatch_time;

/*
 * This is a dispatch table to simplify error checking,
 * and supply return attributes for NFS functions.
 */

#ifdef __STDC__
#define CONCAT(a,b)	a##b
#define CONCAT3(a,b,c)	a##b##c
#define STRING(a)	#a
#else
#define CONCAT(a,b)	a/**/b
#define CONCAT3(a,b,c)	a/**/b/**/c
#define STRING(a)	"a"
#endif

#define table_ent(res_type, arg_type, funct) {			\
	sizeof(res_type), sizeof(arg_type),			\
	(xdrproc_t) CONCAT(xdr_,res_type),			\
	(xdrproc_t) CONCAT(xdr_,arg_type),			\
	(int (*)()) CONCAT3(nfsd_nfsproc_,funct,_2),		\
	STRING(funct), CONCAT(pr_,arg_type)			\
}

#ifdef NFSV3
#define table_ent3(res_type, arg_type, funct) {			\
	sizeof(res_type), sizeof(arg_type),			\
	(xdrproc_t) CONCAT(xdr_,res_type),			\
	(xdrproc_t) CONCAT(xdr_,arg_type),			\
	(int (*)()) CONCAT3(nfsd_nfsproc_,funct,_3),		\
	STRING(funct), CONCAT(pr_,arg_type)		\
}
#endif /* NFSV3 */

/* This definition makes the first table entry use xdr_char instead of
 * xdr_void. The NULL procedure call over TCP fails for this reason
#define nil	char
 */
#define nil	void
#define xdr_nil	xdr_void
#define pr_nil	pr_void
#define pr_char	pr_void

struct dispatch_entry {
  int		res_size, arg_size;	/* sizeof the res/arg structs	*/
  xdrproc_t	xdr_result;
  xdrproc_t	xdr_argument;
  int		(*funct)();		/* function handler		*/
  char	*name;			/* name of function		*/
  char	*(*log_print)();	/* ptr to debug handler		*/
};

/* Functions for debuggind output - these are hooks, not in use yet */
static char *pr_void(void);
static char *pr_nfs_fh(nfs_fh *argp);
static char *pr_sattrargs(sattrargs *argp);
static char *pr_diropargs(diropargs *argp);
static char *pr_readargs(readargs *argp);
static char *pr_writeargs(writeargs *argp);
static char *pr_createargs(createargs *argp);
static char *pr_renameargs(renameargs *argp);
static char *pr_linkargs(linkargs *argp);
static char *pr_symlinkargs(symlinkargs *argp);
static char *pr_readdirargs(readdirargs *argp);

#ifdef NFSV3
static char *pr_nfs_fh3(nfs_fh3 *argp);
static char *pr_setattr3args(setattr3args *argp);
static char *pr_diropargs3(diropargs3 *argp);
static char *pr_access3args(access3args *argp);
static char *pr_read3args(read3args *argp);
static char *pr_write3args(write3args *argp);
static char *pr_create3args(create3args *argp);
static char *pr_mkdir3args(mkdir3args *argp);
static char *pr_symlink3args(symlink3args *argp);
static char *pr_mknod3args(mknod3args *argp);
static char *pr_rename3args(rename3args *argp);
static char *pr_link3args(link3args *argp);
static char *pr_readdir3args(readdir3args *argp);
static char *pr_readdirplus3args(readdirplus3args *argp);
static char *pr_commit3args(commit3args *argp);
#endif /* NFSV3 */

static struct dispatch_entry dtable[] = {
  table_ent(nil,nil,null),			/* NULL */
  table_ent(attrstat,nfs_fh,getattr),		/* GETATTR */
  table_ent(attrstat,sattrargs,setattr),	/* SETATTR */
  table_ent(nil,nil,root),			/* ROOT */
  table_ent(diropres,diropargs,lookup),		/* LOOKUP */
  table_ent(readlinkres,nfs_fh,readlink),	/* READLINK */
  table_ent(readres,readargs,read),		/* READ */
  table_ent(nil,nil,writecache),		/* WRITECACHE */
  table_ent(attrstat,writeargs,write),		/* WRITE */
  table_ent(diropres,createargs,create),	/* CREATE */
  table_ent(nfsstat,diropargs,remove),		/* REMOVE */
  table_ent(nfsstat,renameargs,rename),		/* RENAME */
  table_ent(nfsstat,linkargs,link),		/* LINK */
  table_ent(nfsstat,symlinkargs,symlink),	/* SYMLINK */
  table_ent(diropres,createargs,mkdir),		/* MKDIR */
  table_ent(nfsstat,diropargs,rmdir),		/* RMDIR */
  table_ent(readdirres,readdirargs,readdir),	/* READDIR */
  table_ent(statfsres,nfs_fh,statfs),		/* STATFS */
};

#ifdef NFSV3
static struct dispatch_entry dtable3[] = {
  table_ent3(nil,nil,null),			/* 0 NULL */
  table_ent3(getattr3res,nfs_fh3,getattr),	/* 1 GETATTR */
  table_ent3(wccstat3,setattr3args,setattr),	/* 2 SETATTR */
  table_ent3(lookup3res,diropargs3,lookup),	/* 3 LOOKUP */
  table_ent3(access3res,access3args,access),	/* 4 ACCESS */
  table_ent3(readlink3res,nfs_fh3,readlink),	/* 5 READLINK */
  table_ent3(read3res,read3args,read),		/* 6 READ */
  table_ent3(write3res,write3args,write),	/* 7 WRITE */
  table_ent3(diropres3,create3args,create),	/* 8 CREATE */
  table_ent3(diropres3,mkdir3args,mkdir),	/* 9 MKDIR */
  table_ent3(diropres3,symlink3args,symlink),	/* 10 SYMLINK */
  table_ent3(diropres3,mknod3args,mknod),	/* 11 MKNOD */
  table_ent3(wccstat3,diropargs3,remove),	/* 12 REMOVE */
  table_ent3(wccstat3,diropargs3,rmdir),	/* 13 RMDIR */
  table_ent3(rename3res,rename3args,rename),	/* 14 RENAME */
  table_ent3(link3res,link3args,link),		/* 15 LINK */
  table_ent3(readdir3res,readdir3args,readdir),	/* 16 READDIR */
  table_ent3(readdirplus3res,readdirplus3args,readdirplus),/* 17 READDIRPLUS */
  table_ent3(fsstat3res,nfs_fh3,fsstat),	/* 18 FSSTAT */
  table_ent3(fsinfo3res,nfs_fh3,fsinfo),	/* 19 FSINFO */
  table_ent3(pathconf3res,nfs_fh3,pathconf),	/* 20 PATHCONF */
  table_ent3(commit3res,commit3args,commit),	/* 21 COMMIT */
};
#endif /* NFSV3 */

/* Profiling data structures and constants */
#ifdef CALL_PROFILING
/* time in which proxy was started */
struct timeval   inittime;
/* t0, t1: time in which an NFS request begins/ends processing */
static struct timeval	t0, t1;

/* These structures record: processing time, number of calls,
   and request size for each of the possible V2/V3 calls (22 total) */
static struct timeval	rtimes[22] = {
  {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
  {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
  {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
  {0, 0}, {0, 0}, {0, 0}, {0, 0}
};
static int		calls[22] = {
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};
static int              size[22] = {
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};
static int              cache_hits[22] = {
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

void log_cache_hits(int proc_index) {
  cache_hits[proc_index]++;
}
#endif /* CALL_PROFILING */

extern void dump_attr_cache_parameters(FILE *fp);
extern void dump_data_cache_parameters(FILE *fp);
extern int data_attr_cache_hit;

/****************************************************************************
 * The main dispatch routine, called every time an RPC request is
 * received by the proxy
 ****************************************************************************/
void nfs_dispatch(struct svc_req *rqstp, SVCXPRT *transp)
{
  unsigned int prog_index = rqstp->rq_vers;
  unsigned int proc_index = rqstp->rq_proc;
  struct dispatch_entry *dent;

  /* Make sure SIGHUP does not update configuration while
     processing an NFS request */
  _rpcsvcdirty = 1;

  /* check sanity of NFS request: program number and procedure
     number */
#ifdef NFSV3 
  if (prog_index != nfsvers) {
    svcerr_progvers(transp,nfsvers,nfsvers);
    goto done;
  }
  if (prog_index == 3) {
    if (proc_index >= (sizeof(dtable3) / sizeof(dtable3[0]))) {
      svcerr_noproc(transp);
      goto done;
    }
    dent = &dtable3[proc_index];
  } else 
#endif /* NFSV3 */
    {
      if (proc_index >= (sizeof(dtable) / sizeof(dtable[0]))) {
	svcerr_noproc(transp);
	goto done;
      }
      dent = &dtable[proc_index];
    }

#ifdef CALL_PROFILING
  gettimeofday(&t0, NULL);
#endif

  /* Initialize our variables for determining the attributes of
   * the file system in nfsd.c */
  nfsclient = NULL;

  /* Decode arguments of NFS call into &argument */
  memset(&argument, 0, dent->arg_size);
  if (!svc_getargs(transp, (xdrproc_t) dent->xdr_argument,
		   (caddr_t) &argument)) {
    Dprintf(L_ERROR, "nfs_dispatch: cannot decode argument\n");    
    svcerr_decode(transp);
    goto done;
  }

  /* Clear out the result structure. */
  memset(&result, 0, dent->res_size);

  /* Log the call. The if() saves us a superfluous call to the debug
   * functions, which is a big performance win.
   */
  if (logging_enabled(D_CALL)) {
    Dprintf(D_CALL,"prog_index = %i, proc_index=%i\n",prog_index,proc_index);
    log_call(rqstp, dent->name, dent->log_print(&argument));
  }

  /* reset cache hit information */
  data_attr_cache_hit = 0;

  /* Do the function call itself: send &argumen, get &result.
     This involves forwarding the request, and mapping return arguments. */
  nfs_dispatch_time = time(NULL);
  result.nfsstat = (*dent->funct) (&argument, rqstp);
  Dprintf(D_CALL, "result: %d\n", result.nfsstat);

  /* Send response back to client */
  svc_sendreply(transp, dent->xdr_result, (caddr_t) &result);

  /* Free up RPC XDR result data structure */
  /* only necessary if not a cache hit (request was actually
     forwarded) */
  if(!data_attr_cache_hit) {
    switch(prog_index) {
    case 2 :
      if (!clnt_freeres(nfsdclient2, (xdrproc_t) dent->xdr_result, 
			(caddr_t) &result)) {
	Dprintf(L_ERROR, "unable to free RPC results, exiting\n");
	exit(-1);
      }
      break;
#ifdef NFSV3
    case 3 :
      if (!clnt_freeres(nfsdclient3, (xdrproc_t) dent->xdr_result, 
			(caddr_t) &result)) {
	Dprintf(L_ERROR, "unable to free RPC results, exiting\n");
	exit(-1);
      }
      break;
#endif /* NFSV3 */
    }
  }

#ifdef CALL_PROFILING
  if(prog_index==2) { /* NFS-V2 */
    if(proc_index==NFSPROC_READ) { /* count bytes read */
      size[proc_index]+=((readargs *)&argument)->count;
    }
    if(proc_index==NFSPROC_WRITE) { /* count bytes written */
      size[proc_index]+=((writeargs *)&argument)->totalcount;
    }
  }
#ifdef NFSV3
  if(prog_index==3) { /* NFS-V3 */
    if(proc_index==NFSPROC3_READ) { /* count bytes read */ 
      size[proc_index]+=((read3args *)&argument)->count;
    }
    if(proc_index==NFSPROC3_WRITE) { /* count bytes written */
      size[proc_index]+=((write3args *)&argument)->count;
    } 
  }
#endif /* NFSV3 */
#endif /* CALL_PROFILING */

  /* Free up RPC argument data structure */
  if (!svc_freeargs(transp, (xdrproc_t) dent->xdr_argument,
		    (caddr_t) &argument)) {
    Dprintf(L_ERROR, "nfs_dispatch: unable to free RPC arguments, exiting\n");
    exit(-1);
  }

#ifdef CALL_PROFILING
  /* record time it took to process request */
  gettimeofday(&t1, NULL);
  if (t1.tv_usec < t0.tv_usec) {
    rtimes[proc_index].tv_sec += t1.tv_sec - t0.tv_sec - 1;
    rtimes[proc_index].tv_usec += 1000000 + t1.tv_usec - t0.tv_usec;
  } else {
    rtimes[proc_index].tv_sec += t1.tv_sec - t0.tv_sec;
    rtimes[proc_index].tv_usec += t1.tv_usec - t0.tv_usec;
  }
  calls[proc_index]++;
#endif

done:
  _rpcsvcdirty = 0;
  if (need_reinit) {
    /* in case a SIGHUP was received while dispatching */
    reinitialize(0);
  }
  if (need_flush) {
    flush_caches(0);
  }
  if (need_alarm) {
    setalarm(0);
  }
}



/****************************************************************************
 * Dump profile statistics to a file
 ****************************************************************************/
#ifdef CALL_PROFILING
void dump_stats(int sig)
{
  char PATH_PROFILE[100];
  int     totalcalls=0;
  FILE	*fp;
  int	i;

  Dprintf(L_WARNING, "Call profile requested\n");
  sprintf(PATH_PROFILE,"/tmp/gvfs.nfsd-profile%i",(int)getpid());
  if ((fp = fopen(PATH_PROFILE, "w")) == NULL) {
    Dprintf(L_WARNING, "unable to write profile data to %s\n",
	    PATH_PROFILE);
    return;
  }

  fprintf(fp,"Profile of NFS calls\n");
  dump_attr_cache_parameters(fp);
  dump_data_cache_parameters(fp);

#ifdef NFSV3
  for (i = 0; i < 22; i++) {
#else /* !NFSV3 */
  for (i = 0; i < 18; i++) {
#endif /* NFSV3 */
    float t;
    t = (float) rtimes[i].tv_sec +
      (float) rtimes[i].tv_usec / 1000000.0;
#ifdef NFSV3
    fprintf(fp,"RPC FUNCTION NAME: %s(V3), %s(V2)\n", dtable3[i].name,
	    (i<18 ? dtable[i].name : "n/a"));
#else
    fprintf(fp,"RPC FUNCTION NAME: %s\n", dtable[i].name);
#endif /* NFSV3 */
    fprintf(fp,"\tNumber of calls:\t\t\t%i\n",calls[i]);
    fprintf(fp,"\tCache hits:\t\t\t\t%i\n",cache_hits[i]);
    fprintf(fp,"\tCache hit rate:\t\t\t\t%.3f\n",
	    (calls[i]==0 ? 0.0 : 
	     ((float)cache_hits[i])/((float)calls[i])));
    fprintf(fp,"\tTotal bytes transferred (r/w):\t\t%i\n",
	    size[i]);
    fprintf(fp,"\tTotal time servicing requests (s):\t%.3e\n",t);
    fprintf(fp,"\tAverage time servicing a request (s):\t%.3e\n",
	    (calls[i])?  t / calls[i] : 0);
    fprintf(fp,"\tAverage bandwidth (B/s):\t\t%.3e\n",
	    (t==0.0)? 0: size[i]/t);

    rtimes[i].tv_sec = rtimes[i].tv_usec = 0;
    totalcalls+=calls[i];
    calls[i] = 0;
    size[i] = 0;
    cache_hits[i] = 0;
  }

  fprintf(fp, "\nTOTAL_CALLS: %i\n",totalcalls);
  gettimeofday(&t1, NULL);
  fprintf(fp, "TOTAL_TIME: %i\n", (int) t1.tv_sec - 
	  (int) inittime.tv_sec);
  fclose (fp);
}
#endif /* CALL_PROFILING */


/****************************************************************************
 * Functions for debugging output - these are hooks, not in use yet.
 ****************************************************************************/
char *fh_pr(nfs_fh *fh)
{
  return NULL;
}

static char *pr_void(void)
{
	return ("");
}

static char *pr_nfs_fh(nfs_fh *argp)
{
	return ("");
}

static char *pr_sattrargs(sattrargs *argp)
{
	return ("");
}

static char *pr_diropargs(diropargs *argp)
{
	return ("");
}

static char *pr_readargs(readargs *argp)
{
	return ("");
}

static char *pr_writeargs(writeargs *argp)
{
	return ("");
}

static char *pr_createargs(createargs *argp)
{
	return ("");
}

static char *pr_renameargs(renameargs *argp)
{
	return ("");
}

static char *pr_linkargs(linkargs *argp)
{
	return ("");
}

static char *pr_symlinkargs(symlinkargs *argp)
{
	return ("");
}

static char *pr_readdirargs(readdirargs *argp)
{
	return ("");
}

#ifdef NFSV3
static char *pr_nfs_fh3(nfs_fh3 *argp)
{
	return ("");
}

static char *pr_setattr3args(setattr3args *argp)
{
	return ("");
}

static char *pr_diropargs3(diropargs3 *argp)
{
	return ("");
}

static char *pr_access3args(access3args *argp)
{
	return ("");
}

static char *pr_read3args(read3args *argp)
{
	return ("");
}

static char *pr_write3args(write3args *argp)
{
	return ("");
}

static char *pr_create3args(create3args *argp)
{
	return ("");
}

static char *pr_mkdir3args(mkdir3args *argp)
{
	return ("");
}

static char *pr_symlink3args(symlink3args *argp)
{
	return ("");
}

static char *pr_mknod3args(mknod3args *argp)
{
	return ("");
}

static char *pr_rename3args(rename3args *argp)
{
	return ("");
}

static char *pr_link3args(link3args *argp)
{
	return ("");
}

static char *pr_readdir3args(readdir3args *argp)
{
	return ("");
}

static char *pr_readdirplus3args(readdirplus3args *argp)
{
	return ("");
}

static char *pr_commit3args(commit3args *argp)
{
	return ("");
}
#endif /* NFSV3 */
