/*
 * Copyright (c) 2002-2005, In-VIGO Team
 * Advanced Computing and Information Systems (ACIS) Laboratory
 * University of Florida
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * nfsd_clnt.c		This file has funtions that handle the
 *                      creation of the client-side of the gvfs proxy
 *
 * Author: Renato J. Figueiredo
 *         Ming Zhao
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#include <rpc/pmap_clnt.h>
#ifdef SYSV
#include <rpc/clnt_soc.h>
#endif
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "nfs_prot.h"
#include "nfsd.h"

#ifdef SOLARIS
#define PORTMAP
#include <rpc/rpc.h>
#include <rpc/clnt_soc.h>
#endif /* SOLARIS */

/* default values for client timeouts and retry intervals */
struct timeval timeout = { 180, 0 };
struct timeval retry_timeout = { 20, 0 };




/****************************************************************************
 * Setup a connection to host "svr", program "prog" and version "vers"
 * using a pre-opened privileged port (in case rootsock!=0) or any
 * port (in case rootsock==0). Return a socket ID.
 ****************************************************************************/
int setup(int type, struct sockaddr_in *svr, int prog, int vers, int rootsock)
{
  int s;

  if(rootsock==0) {
    s = RPC_ANYSOCK;
  } else { /* privileged port (rootsock) has been inherited */
    s = rootsock;
  }
  return s;
}




/****************************************************************************
 * Returns an auth handle for the client, with uid/gid same as
 * of proxy's owner.
 ****************************************************************************/
AUTH *create_authenticator(int authtype)
{
  if (authtype == AUTH_UNIX) {
    return authunix_create_default();
  } else {
    fprintf(stderr, "create_authenticator: no secure nfs support\n");
    return(NULL);
  }
}

/****************************************************************************
 * Returns an auth handle for the client, with the option of
 * overriding uid/gid credentials.
 ***************************************************************************/

AUTH *create_authenticator_with_map(int authtype, int uid, int gid)
{
  gid_t gids[1];
  char machname[1024];

  if (authtype == AUTH_UNIX) {
    gids[0] = gid;
    if((gethostname(machname,1024))==-1) {
      Dprintf(L_ERROR, "create authenticator: cannot obtain host name\n");
      return(NULL);
    }
    return authunix_create(machname, uid, gid, 1, gids);
  } else {
    Dprintf(L_ERROR, "create_authenticator: no secure nfs support\n");
    return(NULL);
  }
}



/****************************************************************************
 * Connect the client to an NFS server
 * nfshost : the server's address
 * nfsport : the port it listens to
 * vers    : the version of the NFS protocol to connect to
 * rootsock: ordinarily, zero; non-zero if there is a pre-opened
 *           privileged socket in which the connection should be
 *           established (rootsock is the ID of the preopened
 *           socket in this case)
 ****************************************************************************/
CLIENT *open_nfsdclient(char *nfshost, int nfsport, int vers, 
			int rootsock, int flags)
{
  int proto, sock;
  struct sockaddr_in nfsserver_addr; /* remote nfs server address */
  CLIENT *retclnt = NULL;
  struct hostent *hp;

  /* handle numeric and string IP addresses */
  if (inet_addr(nfshost)!=-1) { /* numeric; -1 = INADDR_NONE */
    nfsserver_addr.sin_addr.s_addr = inet_addr(nfshost);
  } else { /* string */
    hp = gethostbyname(nfshost);
    if (hp == NULL) {
      fprintf(stderr, "%s: unknown host\n", nfshost);
      return NULL;
    }
    memcpy(&nfsserver_addr.sin_addr.s_addr, hp->h_addr, hp->h_length);
    nfshost = (char *) hp->h_name;
  }
  
  nfsserver_addr.sin_family = AF_INET;
  nfsserver_addr.sin_port = ntohs(nfsport);

  /* Set up a UDP/TCP proxy NFS client and connected to the NFS server
     If successful, retclnt is a pointer to the NFS client */
  switch (flags & TRANSPORT_MASK) {
  case NFS_OVER_UDP:
    Dprintf(D_CALL,"nfsd_client: opening UDP connection to %s\n", nfshost);
    /* try using NFS over UDP (standard Sun setup) */
    proto = IPPROTO_UDP;
    sock = setup(SOCK_DGRAM, &nfsserver_addr, NFS_PROGRAM, vers, rootsock);
    if ((retclnt = clntudp_bufcreate(&nfsserver_addr, NFS_PROGRAM, vers, 
				     timeout, &sock, NFS_MAXDATA,
				     NFS_MAXDATA)) == (CLIENT *)0) {
      clnt_pcreateerror("nfs clntudp_create");
      if (sock != RPC_ANYSOCK)
	close(sock);
      return NULL;
    }
    break;
  case NFS_OVER_TCP:
    Dprintf(D_CALL,"nfsd_client: opening TCP connection to %s\n", nfshost);
    /* try using NFS over TCP (standard Sun setup) */
    proto = IPPROTO_TCP;
    sock = setup(SOCK_STREAM, &nfsserver_addr, NFS_PROGRAM, vers, rootsock);
    if ((retclnt = clnttcp_create(&nfsserver_addr, NFS_PROGRAM, vers,
				  &sock, NFS_MAXDATA, NFS_MAXDATA)
	 ) == (CLIENT *)0) {
      clnt_pcreateerror("nfs clnttcp_create");
      if (sock != RPC_ANYSOCK)
	close(sock);
      return NULL;
    }
    break;
  default:
    Dprintf(L_FATAL,"nfsd_client: unknwon transport\n");
    break;
  }

  /* set timeouts */
  clnt_control(retclnt, CLSET_TIMEOUT, (char *)&timeout);
  clnt_control(retclnt, CLSET_RETRY_TIMEOUT, (char *)&retry_timeout);

  /* set authentication credentials */
  retclnt->cl_auth = create_authenticator(AUTH_UNIX);
  if(retclnt -> cl_auth==NULL)
    return NULL;

  /* return client */
  return retclnt;
}



/****************************************************************************
 * Close proxy NFS client
 ****************************************************************************/
void
close_nfsdclient(CLIENT *retclnt)
{
  if (retclnt) {
    auth_destroy(retclnt->cl_auth);
    clnt_destroy(retclnt);
  }
}
