/* UNFSD - copyright Mark A Shand, May 1988.
 * This software maybe be used for any purpose provided
 * the above copyright notice is retained.  It is supplied
 * as is, with no warranty expressed or implied.
 *
 * Redone from ground up by Olaf Kirch, April 1995.
 *
 * Rewritten again October 1997, Olaf Kirch <okir@monad.swb.de>.
 *
 * TODO: 
 *  -	time out uids/gids.
 *  -	Write protocol version 2 to allow bulk transfers and
 *	some more intelligent form of authentication.
 *
 *	Authors:
 *		Mark A. Shand
 *		Olaf Kirch, <okir@monad.swb.de>
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <values.h>
#include <pwd.h>
#include <grp.h>
#include <fcntl.h>
#include "nfsd.h"
#include "ugid.h"

#ifdef SOLARIS
#define PORTMAP
#include <rpc/rpc.h>
#include <rpc/svc_soc.h>
#endif /* SOLARIS */

#define UGID_CHUNK		256
#define UGID_CHUNK_BITS		8
#define UGID_CHUNK_BYTES	(UGID_CHUNK * sizeof(idmap_t *))
#define UGID_CHUNK0_BYTES	(UGID_CHUNK * sizeof(idmap_t))
#define UGID_EXPIRE		300		/* 5 minutes */

/*
 * Make sure we get the right size for ugid_t
 */
#if SIZEOF_UID_T != SIZEOF_GID_T
#error Sorry, this code relies on sizeof(uid_t) == sizeof(gid_t)
#endif

#define UGID_BITS		(SIZEOF_UID_T * BITSPERBYTE)
#define BITSTOLEVEL(b)		((UGID_BITS - (b)) / UGID_CHUNK_BITS - 1)
#define UGID_LOWER(id, b)	((id) & ~((1 << (b)) - 1))
#define UGID_UPPER(id, b)	(((id + (1 << (b))) & ~((1 << (b)) - 1))-1)

/*
 * Common type for uids/gids
 */
typedef uid_t			ugid_t;
typedef struct {
	ugid_t			id;
	time_t			expire;
} idmap_t;

/*
 * This struct holds the entire uid/gid mapping.
 * Note that we don't really keep the mapping in a huge consecutive
 * list, but rather in a multi-level array.
 * See ugid_get_entry for details.
 */
typedef struct ugid_map {
	idmap_t **		map[4];
} ugid_map;

/*
 * uid/gid map indices in ugid_map.
 * Don't change the numbering unless you know what you're doing.
 */
#define MAP_UID_R2L		0
#define MAP_UID_L2R		1
#define MAP_GID_R2L		2
#define MAP_GID_L2R		3
#define MAP_REVERSE(how)	((how) ^ 1)

/*
 * Check whether requested mapping flavor is dynamic
 */
#define MAP_DYNAMIC(map)	((map) == map_ugidd || (map) == map_nis)


/*
 * Prototypes and the like
 */
static ugid_map *	ugid_get_map(nfs_mount *mountp);
static idmap_t *	ugid_get_entry(idmap_t **, ugid_t, int);
static int		rlookup(nfs_mount *, struct svc_req *, int,
					ugid_t, idmap_t *);

/*
 * Define a static mapping
 */
static inline void
ugid_map_static(idmap_t **map, ugid_t from, ugid_t to)
{
  idmap_t		*ent;

  ent = ugid_get_entry(map, from, 1);
  ent->id = to;
  ent->expire = 0;
}

/*
 * Define a dynamic mapping
 */
static inline void
ugid_map_dynamic(idmap_t **map, ugid_t from, ugid_t to)
{
  idmap_t		*ent;

  ent = ugid_get_entry(map, from, 1);
  ent->id = to;
  ent->expire = nfs_dispatch_time + UGID_EXPIRE;
}

/*
 * Find the corresponding id.
 */
static inline ugid_t
ugid_find(nfs_mount *mountp, struct svc_req *rqstp,
			int how, ugid_t id, ugid_t anonid)
{
  ugid_map	*umap;
  idmap_t		*ent;
  
  umap = ugid_get_map(mountp);
  
  if (mountp->o.uidmap == map_static) {
    ent = ugid_get_entry(umap->map[how], id, 0);
    if (ent == 0
	|| ent->id == AUTH_UID_NONE
	|| ent->id == AUTH_UID_NOBODY)
      return anonid;
    return ent->id;
  }
  
  if (mountp->o.uidmap == identity) {
    ent = ugid_get_entry(umap->map[how], id, 0);
    if (ent == 0 || ent->id == AUTH_UID_NONE)
      return id;
    if (ent->id == AUTH_UID_NOBODY)
      return anonid;
    return ent->id;
  }
  
  /* Dynamic mapping flavors */
  ent = ugid_get_entry(umap->map[how], id, 1);
  if (ent->id == AUTH_UID_NONE) {
    rlookup(mountp, rqstp, how, id, ent);
    if (ent->id == AUTH_UID_NONE) {
      ent->id = anonid;
    } else {
      /* Create a dynamic entry in the reverse map */
      ugid_map_dynamic(umap->map[MAP_REVERSE(how)],
		       ent->id, id);
    }
  }
  
  return ent->id;
}

/*
 * Map a server uid to a client uid
 */
uid_t
ruid(uid_t uid, nfs_mount *mountp, struct svc_req *rqstp)
{
  uid_t		retuid;

#ifdef DOSHACKS
  /* Reverse effects of all_squash for DOS clients */
  if (mountp->o.all_squash && uid == mountp->o.nobody_uid)
    return cred_uid;
#endif
  if (mountp->o.uidmap == identity)
    return uid;

  retuid = ugid_find(mountp, rqstp, MAP_UID_L2R, uid, mountp->o.nobody_uid);
  
  Dprintf(D_UGID, "ruid(%s, %d) = %d\n",
	  inet_ntoa(mountp->client->clnt_addr), uid, retuid);
  return retuid;
}

/*
 * Map a server gid to a client gid.
 */
gid_t
rgid(gid_t gid, nfs_mount *mountp, struct svc_req *rqstp)
{
  gid_t		retgid;

#ifdef DOSHACKS
  /* Reverse effects of all_squash for DOS clients */
  if (mountp->o.all_squash && gid == mountp->o.nobody_gid)
    return cred_gid;
#endif
  if (mountp->o.uidmap == identity)
    return gid;
  
  retgid = ugid_find(mountp, rqstp, MAP_GID_L2R, gid, mountp->o.nobody_gid);
  
  Dprintf(D_UGID, "rgid(%s, %d) = %d\n",
	  inet_ntoa(mountp->client->clnt_addr), gid, retgid);
  return retgid;
}

/*
 * Map a client uid to a server uid
 */
uid_t
luid(uid_t uid, nfs_mount *mountp, struct svc_req *rqstp)
{
  uid_t		retuid = uid;

  if (mountp->o.uidmap != identity || mountp->o.some_squash)
    retuid = ugid_find(mountp, rqstp, MAP_UID_R2L, uid,
		       mountp->o.nobody_uid);

  if ((retuid == 0 && mountp->o.root_squash) || mountp->o.all_squash)
    retuid = mountp->o.nobody_uid;

  Dprintf(D_UGID, "luid(%s, %d) = %d\n",
	  inet_ntoa(mountp->client->clnt_addr), uid, retuid);
  return retuid;
}

/*
 * Map a client gid to a server gid
 */
gid_t
lgid(gid_t gid, nfs_mount *mountp, struct svc_req *rqstp)
{
  gid_t		retgid = gid;

  if (mountp->o.uidmap != identity || mountp->o.some_squash)
    retgid = ugid_find(mountp, rqstp, MAP_GID_R2L, gid,
		       mountp->o.nobody_gid);
  
  if ((gid == 0 && mountp->o.root_squash) || mountp->o.all_squash)
    retgid = mountp->o.nobody_gid;
  
  Dprintf(D_UGID, "lgid(%s, %d) = %d\n",
	  inet_ntoa(mountp->client->clnt_addr), gid, retgid);
  return retgid;
}

/*
 * Squash a range of uids/gids.
 */
void
ugid_squash_uids(nfs_mount *mountp, uid_t lo, uid_t hi)
{
  ugid_map	*umap;
  
  Dprintf(D_UGID, "%s:%s squash uids %d-%d\n",
	  mountp->client->clnt_name, mountp->path, lo, hi);
  umap = ugid_get_map(mountp);
  while (lo <= hi)
    ugid_map_static(umap->map[MAP_UID_R2L], lo++, AUTH_UID_NOBODY);
}

/*
 * Squash a range of gids
 */
void
ugid_squash_gids(nfs_mount *mountp, gid_t lo, gid_t hi)
{
  ugid_map	*umap;
  
  Dprintf(D_UGID, "%s:%s squash gids %d-%d\n",
	  mountp->client->clnt_name, mountp->path, lo, hi);
  umap = ugid_get_map(mountp);
  while (lo <= hi)
    ugid_map_static(umap->map[MAP_GID_R2L], lo++, AUTH_GID_NOBODY);
}

/*
 * Define client to server mapping records for a given uid or gid.
 */
void
ugid_map_uid(nfs_mount *mountp, uid_t from, uid_t to)
{
  ugid_map	*umap;
  
  Dprintf(D_UGID, "%s:%s map uid rem %d <-> loc %d\n",
	  mountp->client->clnt_name, mountp->path, from, to);
  umap = ugid_get_map(mountp);
  ugid_map_static(umap->map[MAP_UID_R2L], from, to);
  ugid_map_static(umap->map[MAP_UID_L2R], to, from);
}

void
ugid_map_gid(nfs_mount *mountp, gid_t from, gid_t to)
{
  ugid_map	*umap;
  
  Dprintf(D_UGID, "%s:%s map gid rem %d <-> loc %d\n",
	  mountp->client->clnt_name, mountp->path, from, to);
  umap = ugid_get_map(mountp);
  ugid_map_static(umap->map[MAP_GID_R2L], from, to);
  ugid_map_static(umap->map[MAP_GID_L2R], to, from);
}

/*
 * Get the pointer to a uid/gid map entry.
 */
static idmap_t *
ugid_get_entry(idmap_t **map, ugid_t id, int create)
{
  unsigned int	i, offset, bits = UGID_BITS - UGID_CHUNK_BITS;
  idmap_t		*result, *chunk;
  
  Dprintf(D_UGID, "ugid_get_entry(%p, %d)\n", map, id);
  while (bits > UGID_CHUNK_BITS) {
    offset = (id >> bits) & (UGID_CHUNK - 1);
    if (map[offset] == 0) {
      if (!create)
	return 0;
      chunk = (idmap_t *) xmalloc(UGID_CHUNK_BYTES);
      memset(chunk, 0, UGID_CHUNK * sizeof(*map));
      map[offset] = chunk;
      
      Dprintf(D_UGID, "alloc ptr map %p @ level %d "
	      "(id %u-%u)\n",
	      chunk, BITSTOLEVEL(bits),
	      UGID_LOWER(id, bits),
	      UGID_UPPER(id, bits));
    }
    map   = (idmap_t **) map[offset];
    bits -= UGID_CHUNK_BITS;
  }
  
  offset = (id >> bits) & (UGID_CHUNK - 1);
  if (map[offset] == 0) {
    if (!create)
      return 0;
    chunk = (idmap_t *) xmalloc(UGID_CHUNK0_BYTES);
    Dprintf(D_UGID, "alloc id  map %p @ level %d "
	    "(id %u-%u)\n",
	    chunk, BITSTOLEVEL(bits),
	    UGID_LOWER(id, bits),
	    UGID_UPPER(id, bits));
    map[offset] = chunk;
    
    for (i = 0; i < UGID_CHUNK; i++, chunk++) {
      chunk->id     = AUTH_UID_NONE;
      chunk->expire = 0;	/* never */
    }
  }
  
  result = map[offset] + (id & (UGID_CHUNK - 1));
  if (result->expire && result->expire < nfs_dispatch_time)
    result->id = AUTH_UID_NONE;
  Dprintf(D_UGID, "\tresult = %p\n", map);
  return result;
}

/*
 * Get the map for a given mount point. If it hasn't been initialized yet,
 * create it.
 */
static ugid_map *
ugid_get_map(nfs_mount *mountp)
{
  nfs_client	*clientp = mountp->client;
  struct ugid_map	*umap;
  unsigned int	how;
  
  if (clientp->umap == NULL) {
    clientp->umap = umap = (ugid_map *) xmalloc(sizeof(ugid_map));
    memset(umap, 0, sizeof(ugid_map));
    
    for (how = 0; how < 4; how++) {
      umap->map[how] = (idmap_t **) xmalloc(UGID_CHUNK_BYTES);
      memset(umap->map[how], 0, UGID_CHUNK_BYTES);
    }
  }
  
  return clientp->umap;
}

static void
ugid_do_free_map(idmap_t **map, ugid_t id, unsigned int bits)
{
  unsigned int	i, isptrmap;
  unsigned int	subbits = bits - UGID_CHUNK_BITS;
  
  isptrmap = (bits > UGID_CHUNK_BITS);
  
  isptrmap = (bits > UGID_CHUNK_BITS);
  for (i = 0; i < UGID_CHUNK; i++) {
    if (map[i]) {
      Dprintf(D_UGID, "free  %s map %p @ level %d "
	      "(id %u-%u)\n",
	      isptrmap? "ptr" : "id ",
	      map[i], BITSTOLEVEL(bits),
	      UGID_LOWER(id, bits), UGID_UPPER(id, bits));
      if (isptrmap) {
	ugid_do_free_map((idmap_t **) map[i],
			 id, subbits);
      }
      free(map[i]);
      map[i] = 0;
    }
    id += (1 << bits);
  }
}

/*
 * Deallocate a uid map.
 */
void
ugid_free_map(ugid_map *umap)
{
  unsigned int	how;
  
  /* Free idmap's associated with map */
  for (how = 0; how < 4; how++) {
    idmap_t	**map;
      
    map = (idmap_t **) umap->map[how];
    ugid_do_free_map(map, 0, UGID_BITS - UGID_CHUNK_BITS);
    /* free(map); */
    umap->map[how] = NULL;
  }
  
  /* Free the map itself. */
  free(umap);
}


/*
 * Lookup a given uid or gid by calling the client's ugidd.
 * These are dummy functions used when ugidd/nis support is not compiled
 * in.
 */
static int
rlookup(nfs_mount *mountp, struct svc_req *rqstp,
				int how, ugid_t loc, idmap_t *rem)
{
  if (mountp->o.uidmap != map_daemon
      && mountp->o.uidmap != map_nis)
    return 1;
  return -1;
}
